from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .subdmesh import SubdMesh
from .pattern import Pattern
from .diagram import Diagram
from .formdiagram import FormDiagram
from .forcediagram import ForceDiagram
from .thrustdiagram import ThrustDiagram

__all__ = [
    "SubdMesh",
    "Pattern",
    "Diagram",
    "FormDiagram",
    "ForceDiagram",
    "ThrustDiagram",
]



import rhinoscriptsyntax as rs  # type: ignore
import scriptcontext as sc  # type: ignore

import compas_rhino.conversions
from compas.colors import Color
from compas.geometry import Cylinder
from compas.geometry import Line
from compas.geometry import Vector
from compas.scene.descriptors.color import ColorAttribute
from compas.scene.descriptors.colordict import ColorDictAttribute
from compas_rui.scene import RUIMeshObject
from compas_rv.datastructures import ThrustDiagram
from compas_rv.session import RVSession


class RhinoThrustObject(RUIMeshObject):
    session = RVSession()
    mesh: ThrustDiagram

    vertexcolor = ColorDictAttribute(default=Color.purple())
    edgecolor = ColorDictAttribute(default=Color.purple().darkened(50))
    facecolor = ColorDictAttribute(default=Color.purple().lightened(25))
    freecolor = ColorAttribute(default=Color.purple())
    anchorcolor = ColorAttribute(default=Color.red())
    fixedcolor = ColorAttribute(default=Color.cyan())
    residualcolor = ColorAttribute(default=Color.cyan())
    reactioncolor = ColorAttribute(default=Color.green())
    loadcolor = ColorAttribute(default=Color.green().darkened(50))
    selfweightcolor = ColorAttribute(default=Color.white())
    compressioncolor = ColorAttribute(default=Color.blue())
    tensioncolor = ColorAttribute(default=Color.red())

    def __init__(
        self,
        disjoint=True,
        show_supports=True,
        show_fixed=True,
        show_free=False,
        vertexgroup="RV::ThrustDiagram::Vertices",
        edgegroup="RV::ThrustDiagram::Edges",
        facegroup="RV::ThrustDiagram::Faces",
        loadgroup="RV::ThrustDiagram::Loads",
        selfweightgroup="RV::ThrustDiagram::Selfweight",
        forcegroup="RV::ThrustDiagram::Forces",
        reactiongroup="RV::ThrustDiagram::Reactions",
        residualgroup="RV::ThrustDiagram::Residuals",
        **kwargs,
    ):
        super().__init__(
            disjoint=disjoint,
            vertexgroup=vertexgroup,
            edgegroup=edgegroup,
            facegroup=facegroup,
            **kwargs,
        )

        self.show_faces = True
        self.show_edges = False
        self.show_supports = show_supports
        self.show_fixed = show_fixed
        self.show_free = show_free
        self.loadgroup = loadgroup
        self.selfweightgroup = selfweightgroup
        self.forcegroup = forcegroup
        self.reactiongroup = reactiongroup
        self.residualgroup = residualgroup

    @property
    def settings(self):
        settings = super().settings
        settings["show_supports"] = self.show_supports
        settings["show_fixed"] = self.show_fixed
        settings["show_free"] = self.show_free
        return settings

    @property
    def diagram(self) -> ThrustDiagram:
        return self.mesh

    @diagram.setter
    def diagram(self, diagram: ThrustDiagram) -> None:
        self.mesh = diagram

    def compute_pipe_colors(self, tol=1e-3) -> None:
        edges = list(self.diagram.edges())
        forces = [self.diagram.edge_attribute(edge, "_f") for edge in edges]
        magnitudes = [abs(f) for f in forces]
        fmin = min(magnitudes)
        fmax = max(magnitudes)

        if fmax - fmin < tol:
            # the size of the range is already checked here
            # no need to do this again in the loop
            return

        colors = []
        for force, magnitude in zip(forces, magnitudes):
            # this will need to be updated when we include tension edges
            colors.append(Color.from_i((magnitude - fmin) / (fmax - fmin)))

        return dict(zip(edges, colors))

    # =============================================================================
    # Clear
    # =============================================================================

    # =============================================================================
    # Draw
    # =============================================================================

    def draw(self):
        faces = []
        if self.show_faces:
            faces += list(self.diagram.faces_where(_is_loaded=True))
        if faces:
            self.show_faces = faces

        for vertex in self.diagram.vertices():
            if self.diagram.vertex_attribute(vertex, "is_support"):
                self.vertexcolor[vertex] = self.anchorcolor
            elif self.diagram.vertex_attribute(vertex, "is_fixed"):
                self.vertexcolor[vertex] = self.fixedcolor
            else:
                self.vertexcolor[vertex] = self.freecolor

        super().draw()

        if self.session.settings.drawing.show_reactions:
            self.draw_reactions()
        if self.session.settings.drawing.show_loads:
            self.draw_loads()
        if self.session.settings.drawing.show_selfweight:
            self.draw_selfweight()
        if self.session.settings.drawing.show_pipes:
            self.draw_pipes()

        return self.guids

    def draw_vertices(self):
        if self.show_vertices is True:
            vertices = []
            if self.show_free:
                vertices += list(self.diagram.vertices_where(is_support=False, is_fixed=False))
            if self.show_fixed:
                vertices += list(self.diagram.vertices_where(is_fixed=True))
            if self.show_supports:
                vertices += list(self.diagram.vertices_where(is_support=True))
            self.show_vertices = vertices

        for vertex in self.diagram.vertices():
            if self.diagram.vertex_attribute(vertex, "is_support"):
                self.vertexcolor[vertex] = self.anchorcolor
            elif self.diagram.vertex_attribute(vertex, "is_fixed"):
                self.vertexcolor[vertex] = self.fixedcolor
            else:
                self.vertexcolor[vertex] = self.freecolor

        return super().draw_vertices()

    def draw_edges(self):
        if self.show_edges is True:
            edges = list(self.diagram.edges_where(_is_edge=True))
            if edges:
                self.show_edges = edges

        return super().draw_edges()

    def draw_faces(self):
        faces = []
        if self.show_faces:
            faces += list(self.diagram.faces_where(_is_loaded=True))
        if faces:
            self.show_faces = faces

        return super().draw_faces()

    def draw_loads(self):
        guids = []

        scale = self.session.settings.drawing.scale_loads
        color = self.loadcolor
        tol = self.session.settings.drawing.tol_vectors

        for vertex in self.diagram.vertices_where(is_support=False):
            load = self.diagram.vertex_attributes(vertex, ["px", "py", "pz"])

            if load is not None:
                vector = Vector(*load) * scale
                if vector.length > tol:
                    name = "{}.vertex.{}.load".format(self.diagram.name, vertex)
                    attr = self.compile_attributes(name=name, color=color, arrow="start")
                    point = self.diagram.vertex_point(vertex)
                    line = Line.from_point_and_vector(point, vector)
                    guid = sc.doc.Objects.AddLine(compas_rhino.conversions.line_to_rhino(line), attr)
                    guids.append(guid)

        if guids:
            if self.loadgroup:
                self.add_to_group(self.loadgroup, guids)
            elif self.group:
                self.add_to_group(self.group, guids)

        self._guids += guids
        return guids

    def draw_selfweight(self):
        guids = []

        scale = self.session.settings.drawing.scale_selfweight
        color = self.selfweightcolor
        tol = self.session.settings.drawing.tol_vectors

        for vertex in self.diagram.vertices_where(is_support=False):
            thickness = self.diagram.vertex_attribute(vertex, "t")

            if thickness:
                area = self.diagram.vertex_area(vertex)
                weight = area * thickness
                point = self.diagram.vertex_point(vertex)
                vector = Vector(0, 0, -weight * scale)
                if vector.length > tol:
                    line = Line.from_point_and_vector(point, vector)
                    name = "{}.vertex.{}.selfweight".format(self.diagram.name, vertex)
                    attr = self.compile_attributes(name=name, color=color, arrow="end")
                    guid = sc.doc.Objects.AddLine(compas_rhino.conversions.line_to_rhino(line), attr)
                    guids.append(guid)

        if guids:
            if self.selfweightgroup:
                self.add_to_group(self.selfweightgroup, guids)
            elif self.group:
                self.add_to_group(self.group, guids)

        self._guids += guids
        return guids

    def draw_pipes(self):
        guids = []

        scale = self.session.settings.drawing.scale_pipes
        tol = self.session.settings.drawing.tol_pipes

        pipe_colors = self.compute_pipe_colors()

        for edge in self.diagram.edges():
            force = self.diagram.edge_attribute(edge, "_f")

            if force != 0:
                line = self.diagram.edge_line(edge)
                radius = abs(force) * scale

                color = self.compressioncolor
                if self.session.settings.drawing.show_forces:
                    color = pipe_colors[edge]

                if radius > tol:
                    pipe = Cylinder.from_line_and_radius(line, radius)
                    name = "{}.edge.{}.force".format(self.diagram.name, edge)
                    attr = self.compile_attributes(name=name, color=color)
                    guid = sc.doc.Objects.AddBrep(compas_rhino.conversions.cylinder_to_rhino_brep(pipe), attr)
                    guids.append(guid)

        if guids:
            if self.forcegroup:
                self.add_to_group(self.forcegroup, guids)
            elif self.group:
                self.add_to_group(self.group, guids)

        self._guids += guids
        return guids

    def draw_reactions(self):
        guids = []

        scale = self.session.settings.drawing.scale_reactions
        tol = self.session.settings.drawing.tol_vectors

        for vertex in self.diagram.vertices_where(is_support=True):
            residual = Vector(*self.diagram.vertex_attributes(vertex, ["_rx", "_ry", "_rz"]))
            vector = residual * scale

            if vector.length > tol:
                name = "{}.vertex.{}.reaction".format(self.diagram.name, vertex)
                attr = self.compile_attributes(name=name, color=self.reactioncolor, arrow="start")
                point = self.diagram.vertex_point(vertex)
                line = Line.from_point_and_vector(point, vector)
                guid = sc.doc.Objects.AddLine(compas_rhino.conversions.line_to_rhino(line), attr)
                guids.append(guid)

        if guids:
            if self.reactiongroup:
                self.add_to_group(self.reactiongroup, guids)
            elif self.group:
                self.add_to_group(self.group, guids)

        self._guids += guids
        return guids

    def draw_residuals(self):
        guids = []

        scale = self.session.settings.drawing.scale_residuals
        tol = self.session.settings.drawing.tol_vectors

        for vertex in self.diagram.vertices_where(is_support=False):
            residual = Vector(*self.diagram.vertex_attributes(vertex, ["_rx", "_ry", "_rz"]))

            vector = residual * scale
            if vector.length > tol:
                name = "{}.vertex.{}.residual".format(self.diagram.name, vertex)
                attr = self.compile_attributes(name=name, color=self.residualcolor, arrow="end")
                point = self.diagram.vertex_point(vertex)
                line = Line.from_point_and_vector(point, vector)
                guid = sc.doc.Objects.AddLine(compas_rhino.conversions.line_to_rhino(line), attr)
                guids.append(guid)

        if guids:
            if self.residualgroup:
                self.add_to_group(self.residualgroup, guids)
            elif self.group:
                self.add_to_group(self.group, guids)

        self._guids += guids
        return guids

    # =============================================================================
    # Redraw
    # =============================================================================

    def redraw_vertices(self):
        rs.EnableRedraw(False)
        self.clear_vertices()
        self.draw_vertices()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_edges(self):
        rs.EnableRedraw(False)
        self.clear_edges()
        self.draw_edges()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_faces(self):
        rs.EnableRedraw(False)
        self.clear_faces()
        self.draw_faces()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw(self):
        rs.EnableRedraw(False)
        self.clear()
        self.draw()
        rs.EnableRedraw(True)
        rs.Redraw()



from compas.datastructures import Mesh
from compas.geometry import angle_vectors


class Diagram(Mesh):
    """
    Base data structure for form, force and thrust diagrams.
    """

    def edge_loop(self, uv):
        """
        Identify all edges on the same loop as a given edge.
        This implementation is different than the base implmentation in the mesh data structure,
        because of the "unloaded" faces in form, force, and thrust diagrams.

        Parameters
        ----------
        uv : tuple[int, int]
            The identifier of the edge.

        Returns
        -------
        list[tuple[int, int]]
            A list of edge identifiers.

        """
        u, v = uv
        f1, f2 = self.edge_faces(uv)

        if f1 is not None and not self.face_attribute(f1, "_is_loaded"):
            loop = []
            for edge in self.face_halfedges(f1):
                if self.edge_attribute(edge, "_is_edge"):
                    loop.append(edge)

        elif f2 is not None and not self.face_attribute(f2, "_is_loaded"):
            loop = []
            for edge in self.face_halfedges(f2):
                if self.edge_attribute(edge, "_is_edge"):
                    loop.append(edge)

        else:
            forward = super(Diagram, self).halfedge_loop((u, v))
            backward = super(Diagram, self).halfedge_loop((v, u))
            loop = []
            for u, v in forward:
                loop.append((u, v))
                if self.vertex_attribute(v, "is_fixed"):
                    break
            if not self.vertex_attribute(backward[0][0], "is_fixed"):
                for u, v in backward:
                    loop.insert(0, (v, u))
                    if self.vertex_attribute(v, "is_fixed"):
                        break
        return loop

    def collapse_small_edges(self, tol=1e-2):
        """
        Collapse the edges that are shorter than a given length.

        Parameters
        ----------
        tol : float, optional
            The threshold length.
            Edges with a length shorter than this value
            will be collapsed.

        Returns
        -------
        None

        """
        for edge in list(self.edges()):
            if self.has_edge(edge):
                if self.edge_length(edge) < tol:
                    self.collapse_edge(edge, t=0.5, allow_boundary=True)

    def smooth(self, fixed, kmax=10):
        """
        Apply area smoothing to the mesh geometry.

        Parameters
        ----------
        fixed : list[int]
            The vertices to keep fixed.
        kmax : int, optional
            The number of smoothing iterations.

        Returns
        -------
        None

        """
        self.smooth_area(self, fixed=fixed, kmax=kmax)

    def corner_vertices(self, tol=160):
        """
        Identify the corner vertices.

        Parameters
        ----------
        tol : float, optional
            The threshold value for the angle formed between two edges at a vertex
            for it to be considered a corner.
            Vertices with smaller angles are considered a corner.

        Returns
        -------
        list[int]

        """
        vkeys = []
        for key in self.vertices_on_boundary():
            if self.vertex_degree(key) == 2:
                vkeys.append(key)
            else:
                nbrs = []
                for nkey in self.vertex_neighbors(key):
                    if self.is_edge_on_boundary(key, nkey):
                        nbrs.append(nkey)
                u = self.edge_vector(key, nbrs[0])
                v = self.edge_vector(key, nbrs[1])
                if angle_vectors(u, v, deg=True) < tol:
                    vkeys.append(key)
        return vkeys

    def edge_loop_vertices(self, uv):
        """
        Identify all vertices on an edge loop.

        Parameters
        ----------
        uv : tuple[int, int]
            The identifier of the base edge of the loop.

        Returns
        -------
        list[int]

        """
        edges = self.edge_loop(uv)
        if len(edges) == 1:
            return edges[0]
        vertices = [edge[0] for edge in edges]
        if edges[-1][1] != edges[0][0]:
            vertices.append(edges[-1][1])
        return vertices



from compas.geometry import Box
from compas.geometry import angle_vectors_xy
from compas.geometry import bounding_box
from compas.geometry import cross_vectors
from compas_fd.solvers import fd_numpy
from compas_tna.diagrams import ForceDiagram

from .diagram import Diagram
from .formdiagram import FormDiagram  # noqa: F401


class ForceDiagram(Diagram, ForceDiagram):
    """
    Data structure for force diagrams.
    """

    def primal_edge(self, edge):
        """
        Get the corresponding edge in the FormDiagram.

        Parameters
        ----------
        edge : tuple[int, int]
            The identifier of the edge in this diagram.

        Returns
        -------
        tuple[int, int]
            The identifier of the edge in the other/primal diagram.

        Raises
        ------
        KeyError
            If the dual edge does not exist.

        """
        primal: FormDiagram = self.primal

        f1, f2 = edge
        for u, v in primal.face_halfedges(f1):
            if primal.halfedge[v][u] == f2:
                return u, v
        raise KeyError(edge)

    def update_angle_deviations(self):
        """
        Compute the angle deviation with the corresponding edge in the FormDiagram.

        Returns
        -------
        None

        """
        primal: FormDiagram = self.primal

        for edge in self.edges():
            edge_ = self.primal_edge(edge)
            uv = self.edge_vector(edge)
            uv_ = primal.edge_vector(edge_)
            a = angle_vectors_xy(uv, cross_vectors(uv_, (0, 0, 1)), deg=True)
            if primal.edge_attribute(edge_, "_is_tension"):
                a = 180 - a
            self.edge_attribute(edge, "_a", a)
            primal.edge_attribute(edge_, "_a", a)

    def update_position(self, margin: float = 1.3) -> None:
        """Update the position of the diagram based on the size and position of the form diagram.

        Parameters
        ----------
        margin : float, optional
            Multiplication factor for the size of the bounding box of the form diagram in the X direction.

        Returns
        -------
        None

        """
        primal: FormDiagram = self.primal

        bbox_form = Box.from_bounding_box(bounding_box(primal.vertices_attributes("xyz")))
        bbox_force = Box.from_bounding_box(bounding_box(self.vertices_attributes("xyz")))

        y_form = bbox_form.ymin + 0.5 * (bbox_form.ymax - bbox_form.ymin)
        y_force = bbox_force.ymin + 0.5 * (bbox_force.ymax - bbox_force.ymin)
        dx = margin * (bbox_form.xmax - bbox_form.xmin) + (bbox_form.xmin - bbox_force.xmin)
        dy = y_form - y_force

        self.translate([dx, dy, 0])

    def solve_fd(self) -> None:
        """Relax the mesh using the force density method with the curent edge force densities of te corresponding form diagram.

        Returns
        -------
        None

        """
        vertex_index = self.vertex_index()

        xyz = self.vertices_attributes("xyz")
        loads = [[0.0, 0.0, 0.0] for _ in xyz]

        fixed = list(self.vertices_where(is_support=True))
        fixed += list(self.vertices_where(is_fixed=True))
        if len(fixed) < 4:
            fixed += self.vertices_on_boundary()
        fixed = list(set([vertex_index[vertex] for vertex in fixed]))

        edges = list(self.edges())
        qinv = [self.form_edge_attribute(self.primal, edge, name="q") for edge in edges]
        q = [1 / value for value in qinv]

        edges = [(vertex_index[u], vertex_index[v]) for u, v in edges]

        result = fd_numpy(
            vertices=xyz,
            fixed=fixed,
            edges=edges,
            forcedensities=q,
            loads=loads,
        )

        for key in self.vertices():
            index = vertex_index[key]
            self.vertex_attributes(key, "xyz", result.vertices[index])



import System  # type: ignore
from Rhino.Geometry import Line  # type: ignore
from Rhino.Geometry import Point3d  # type: ignore
from System.Collections.Generic import List  # type: ignore

from compas_rhino.conduits import BaseConduit

# this needs to be reorganised
# there is not need to update anything since we don't go via RPC anymore
# the data IS live when probably associated with the conduit


class HorizontalConduit(BaseConduit):
    def __init__(self, lines, **kwargs):
        super(HorizontalConduit, self).__init__(**kwargs)
        self._default_thickness = 1.0
        self._default_color = System.Drawing.Color.FromArgb(255, 255, 255)
        self.lines = lines or []

    def DrawForeground(self, e):
        lines = List[Line](len(self.lines))
        for start, end in self.lines:
            lines.Add(Line(Point3d(start[0], start[1], 0), Point3d(end[0], end[1], 0)))
        e.Display.DrawLines(lines, self._default_color, self._default_thickness)



from compas.colors import Color
from compas.scene.descriptors.colordict import ColorDictAttribute
from compas_rv.datastructures import ForceDiagram
from compas_rv.datastructures import FormDiagram
from compas_rv.session import RVSession

from .diagramobject import RhinoDiagramObject


class RhinoForceObject(RhinoDiagramObject):
    session = RVSession()
    diagram: ForceDiagram

    vertexcolor = ColorDictAttribute(default=Color.blue())
    edgecolor = ColorDictAttribute(default=Color.blue().darkened(50))
    facecolor = ColorDictAttribute(default=Color.blue().lightened(25))

    def __init__(
        self,
        vertexgroup="RV::ForceDiagram::Vertices",
        edgegroup="RV::ForceDiagram::Edges",
        facegroup="RV::ForceDiagram::Faces",
        **kwargs,
    ):
        super().__init__(
            vertexgroup=vertexgroup,
            edgegroup=edgegroup,
            facegroup=facegroup,
            **kwargs,
        )

    # =============================================================================
    # Properties
    # =============================================================================

    def forces(self):
        primal: FormDiagram = self.diagram.primal
        edges = [self.diagram.primal_edge(edge) for edge in self.edges()]
        return primal.edges_attribute("_f", keys=edges)

    # =============================================================================
    # Clear
    # =============================================================================

    # =============================================================================
    # Draw
    # =============================================================================

    # =============================================================================
    # Redraw
    # =============================================================================



import rhinoscriptsyntax as rs  # type: ignore

from compas.scene import Scene
from compas_rhino.scene import RhinoSceneObject
from compas_rv.settings import RVSettings
from compas_session.session import Session


def find_all_by_itemtype(scene: Scene, itemtype) -> list[RhinoSceneObject]:
    sceneobjects = []
    for obj in scene.objects:
        if isinstance(obj.item, itemtype):
            sceneobjects.append(obj)
    return sceneobjects


class RVSession(Session):
    settings: RVSettings  # type: ignore

    def __new__(cls, **kwargs):
        if "name" in kwargs:
            del kwargs["name"]
        return super().__new__(cls, name="RhinoVAULT")

    def __init__(self, **kwargs):
        if "name" in kwargs:
            del kwargs["name"]
        super().__init__(name="RhinoVAULT", settings=RVSettings(), **kwargs)

    def clear(self, clear_scene=True, clear_context=True):
        for sceneobject in self.scene.objects:
            if hasattr(sceneobject, "clear_conduits"):
                sceneobject.clear_conduits()  # type: ignore
        self.scene.clear(clear_scene=clear_scene, clear_context=clear_context)

    def clear_conduits(self):
        for sceneobject in self.scene.objects:
            if hasattr(sceneobject, "clear_conduits"):
                sceneobject.clear_conduits()  # type: ignore

    def find_pattern(self, warn=True):
        from compas_rv.datastructures import Pattern
        from compas_rv.scene import RhinoPatternObject

        form: RhinoPatternObject = self.scene.find_by_itemtype(Pattern)  # type: ignore
        if form:
            return form
        if warn:
            rs.MessageBox("There is no Pattern.", title="Warning")

    def find_formdiagram(self, warn=True):
        from compas_rv.datastructures import FormDiagram
        from compas_rv.scene import RhinoFormObject

        form: RhinoFormObject = self.scene.find_by_itemtype(FormDiagram)  # type: ignore
        if form:
            return form
        if warn:
            rs.MessageBox("There is no FormDiagram.", title="Warning")

    def find_forcediagram(self, warn=True):
        from compas_rv.datastructures import ForceDiagram
        from compas_rv.scene import RhinoForceObject

        force: RhinoForceObject = self.scene.find_by_itemtype(ForceDiagram)  # type: ignore
        if force:
            return force
        if warn:
            rs.MessageBox("There is no ForceDiagram.", title="Warning")

    def find_thrustdiagram(self, warn=True):
        from compas_rv.datastructures import ThrustDiagram
        from compas_rv.scene import RhinoThrustObject

        thrust: RhinoThrustObject = self.scene.find_by_itemtype(ThrustDiagram)  # type: ignore
        if thrust:
            return thrust
        if warn:
            rs.MessageBox("There is no ThrustDiagram.", title="Warning")

    def clear_all_patterns(self, redraw=True):
        from compas_rv.datastructures import Pattern

        for obj in find_all_by_itemtype(self.scene, Pattern):
            obj.clear()
            self.scene.remove(obj)
        if redraw:
            self.scene.redraw()
            rs.Redraw()

    def clear_all_diagrams(self, redraw=True):
        self.clear_all_formdiagrams(redraw=False)
        self.clear_all_forcediagrams(redraw=False)
        self.clear_all_thrustdiagrams(redraw=False)
        if redraw:
            self.scene.redraw()
            rs.Redraw()

    def clear_all_formdiagrams(self, redraw=True):
        from compas_rv.datastructures import FormDiagram

        for obj in find_all_by_itemtype(self.scene, FormDiagram):
            obj.clear()
            self.scene.remove(obj)
        if redraw:
            self.scene.redraw()
            rs.Redraw()

    def clear_all_forcediagrams(self, redraw=True):
        from compas_rv.datastructures import ForceDiagram

        for obj in find_all_by_itemtype(self.scene, ForceDiagram):
            obj.clear()
            self.scene.remove(obj)
        if redraw:
            self.scene.redraw()
            rs.Redraw()

    def clear_all_thrustdiagrams(self, redraw=True):
        from compas_rv.datastructures import ThrustDiagram

        for obj in find_all_by_itemtype(self.scene, ThrustDiagram):
            obj.clear()
            self.scene.remove(obj)
        if redraw:
            self.scene.redraw()
            rs.Redraw()

    def confirm(self, message):
        result = rs.MessageBox(message, buttons=4 | 32 | 256 | 0, title="Confirmation")
        return result == 6

    def warn(self, message):
        return rs.MessageBox(message, title="Warning")



from typing import Annotated
from typing import List
from typing import Tuple

import Rhino  # type: ignore
import System  # type: ignore

from compas.colors import Color
from compas_rhino.conduits import BaseConduit


class EdgesConduit(BaseConduit):
    """Display conduit for CableMesh edges as lines.

    Parameters
    ----------
    xyz : list of list of float
        The vertex coordinates.
    edges : list of tuple of int
        List of pairs of indices into the list of vertex coordinates.

    """

    xyz: List[Annotated[List[float], 3]]
    edges: List[Tuple[int, int]]
    thickness: int

    def __init__(self, xyz, edges, thickness=1, color=None, **kwargs):
        super().__init__(**kwargs)
        self.xyz = xyz
        self.edges = edges
        self.thickness = thickness
        self.color = color or Color(0, 0, 0)

    def PostDrawObjects(self, e):
        color = System.Drawing.Color.FromArgb(*self.color.rgb255)
        lines = []
        for i, j in self.edges:
            sp = self.xyz[i]
            ep = self.xyz[j]
            lines.append(
                Rhino.Geometry.Line(
                    Rhino.Geometry.Point3d(*sp),
                    Rhino.Geometry.Point3d(*ep),
                )
            )

        if lines:
            e.Display.DrawLines(lines, color, self.thickness)



from compas.geometry import Box
from compas.geometry import bounding_box
from compas.geometry import cross_vectors
from compas.geometry import length_vector
from compas.geometry import subtract_vectors

from .formdiagram import FormDiagram


class ThrustDiagram(FormDiagram):
    """Data structure for thrust diagrams."""

    def vertex_tributary_area(self, vertex: int) -> float:
        """
        Compute the tributary area of a vertex taking into account only the loaded faces.

        Parameters
        ----------
        vertex : int
            The vertex identifier.

        Returns
        -------
        float

        """
        area = 0
        p0 = self.vertex_coordinates(vertex)
        for nbr in self.halfedge[vertex]:
            p1 = self.vertex_coordinates(nbr)
            v1 = subtract_vectors(p1, p0)
            fkey = self.halfedge[vertex][nbr]
            if fkey is not None:
                if self.face_attribute(fkey, "_is_loaded"):
                    p2 = self.face_centroid(fkey)
                    v2 = subtract_vectors(p2, p0)
                    area += length_vector(cross_vectors(v1, v2))
            fkey = self.halfedge[nbr][vertex]
            if fkey is not None:
                if self.face_attribute(fkey, "_is_loaded"):
                    p3 = self.face_centroid(fkey)
                    v3 = subtract_vectors(p3, p0)
                    area += length_vector(cross_vectors(v1, v3))
        return 0.25 * area

    def vertex_lumped_stress(self, vertex: int) -> float:
        """
        Compute an approximation of the compressive stress at a vertex.

        Parameters
        ----------
        vertex : int
            The vertex identifier.

        Returns
        -------
        float

        """
        stress = 0
        neighbors = self.vertex_neighbors(vertex)
        count = 0
        for nbr in neighbors:
            edge_area = 0
            edge_thickness = sum(self.vertices_attribute("t", keys=[vertex, nbr])) / 2
            edge_force = self.edge_attribute((vertex, nbr), "_f")

            if abs(edge_force) <= 0:
                continue

            mp = self.edge_midpoint((vertex, nbr))

            f0 = self.halfedge_face((vertex, nbr))
            if f0 is not None:
                if self.face_attribute(f0, "_is_loaded"):
                    f0_c = self.face_center(f0)
                    area = length_vector(subtract_vectors(f0_c, mp)) * edge_thickness
                    if area > 0:
                        edge_area += area
            f1 = self.halfedge_face((nbr, vertex))
            if f1 is not None:
                if self.face_attribute(f1, "_is_loaded"):
                    f1_c = self.face_center(f1)
                    area = length_vector(subtract_vectors(f1_c, mp)) * edge_thickness
                    if area > 0:
                        edge_area += area

            if edge_area > 0:
                stress += edge_force / edge_area
                count += 1

        return stress / count

    def compute_zmax(self) -> float:
        """Compute a suitable value for zmax based on the length of the diagonal of the bounding box of the projection of the diagram in XY.

        Returns
        -------
        float
            Maximum Z coordinate.

        """
        bbox = Box.from_bounding_box(bounding_box(self.vertices_attributes("xyz")))
        diagonal = bbox.points[2] - bbox.points[0]
        return 0.25 * diagonal.length



from pydantic import BaseModel
from pydantic import Field

from compas_session.settings import Settings


class TNASettings(BaseModel):
    horizontal_kmax: int = 100
    horizontal_alpha: float = 100
    horizontal_max_angle: float = Field(default=5.0, title="Maximum Angle Deviation", description="The maximum allowed angle between corresponding edges.")
    horizontal_refreshrate: int = 5

    vertical_kmax: int = 300
    vertical_zmax: float = 4.0


class DrawingSettings(BaseModel):
    show_angles: bool = True
    show_forces: bool = False

    show_reactions: bool = True
    show_residuals: bool = False
    show_pipes: bool = False
    show_loads: bool = False
    show_selfweight: bool = False

    show_thickness: bool = False

    scale_reactions: float = 0.1
    scale_residuals: float = 1.0
    scale_pipes: float = 0.01
    scale_loads: float = 1.0
    scale_selfweight: float = 1.0

    tol_vectors: float = 1e-3
    tol_pipes: float = 1e-2


class RVSettings(Settings):
    autoupdate: bool = False

    tna: TNASettings = TNASettings()
    drawing: DrawingSettings = DrawingSettings()



import numpy as np
import scipy.sparse as sps

from compas.linalg import spsolve_with_known
from compas.matrices import connectivity_matrix
from compas_rv.datastructures import ForceDiagram
from compas_rv.datastructures import FormDiagram
from compas_tna.equilibrium.diagrams import rot90


def update_force_from_form(force: ForceDiagram, form: FormDiagram):
    vertex_index = form.vertex_index()

    xy = np.array(form.vertices_attributes("xy"), dtype=np.float64)
    edges = [[vertex_index[u], vertex_index[v]] for u, v in form.edges_where(_is_edge=True)]
    C: sps.csr_matrix = connectivity_matrix(edges, "csr")
    Q = sps.diags([form.edges_attribute("q", keys=list(form.edges_where(_is_edge=True)))], [0])
    uv = C.dot(xy)

    _vertex_index = force.vertex_index()

    _known = [0]

    _xy = np.array(force.vertices_attributes("xy"), dtype=np.float64)
    _xy[:] = rot90(_xy, +1.0)

    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]

    _C: sps.csr_matrix = connectivity_matrix(_edges, "csr")
    _Ct = _C.transpose()

    _xy = spsolve_with_known(_Ct.dot(_C), _Ct.dot(Q).dot(uv), _xy, _known)
    _xy[:] = rot90(_xy, -1.0)

    for vertex, attr in force.vertices(True):
        index = _vertex_index[vertex]
        attr["x"] = _xy[index, 0]
        attr["y"] = _xy[index, 1]



from itertools import groupby

from compas.datastructures import Mesh
from compas.datastructures.mesh.subdivision import mesh_fast_copy
from compas.tolerance import TOL


class SubdMesh(Mesh):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.default_edge_attributes.update(
            {
                "brep_curve": None,
                "brep_curve_pts": [],
                "brep_curve_dir": None,
            }
        )
        self.default_face_attributes.update(
            {
                "is_quad": False,
                "u_edge": None,
                "v_edge": None,
                "nu": 4,
                "nv": 4,
                "n": 2,
                "brep_face": None,
            }
        )
        self._edge_strips = {}

    # @classmethod
    # def from_guid(cls, guid):
    #     """
    #     Construct a subd mesh from a GUID representing a surface or polysurface in Rhino.

    #     Parameters
    #     ----------
    #     guid : System.Guid
    #         The GUID of the input surface.

    #     Returns
    #     -------
    #     :class:`SubdMesh`

    #     """
    #     import compas_rhino.conversions
    #     import compas_rhino.objects
    #     import Rhino  # type: ignore

    #     rhinosurface = compas_rhino.objects.find_object(guid)
    #     subdmesh = compas_rhino.conversions.surface_to_compas_mesh(rhinosurface, cleanup=False)
    #     gkey_vertex = {TOL.geometric_key(subdmesh.vertex_coordinates(vertex)): vertex for vertex in subdmesh.vertices()}

    #     brep = Rhino.Geometry.Brep.TryConvertBrep(rhinosurface.Geometry)

    #     for brep_face, face in zip(brep.Faces, subdmesh.faces()):
    #         loop = brep_face.OuterLoop
    #         curve = loop.To3dCurve()
    #         segments = curve.Explode()

    #         if len(segments) == 4:
    #             domain_u = brep_face.Domain(0)
    #             domain_v = brep_face.Domain(1)
    #             u0_xyz = brep_face.PointAt(domain_u[0], domain_v[0])
    #             u1_xyz = brep_face.PointAt(domain_u[1], domain_v[0])
    #             v0_xyz = brep_face.PointAt(domain_u[0], domain_v[0])
    #             v1_xyz = brep_face.PointAt(domain_u[0], domain_v[1])
    #             u0 = gkey_vertex[TOL.geometric_key(u0_xyz)]
    #             u1 = gkey_vertex[TOL.geometric_key(u1_xyz)]
    #             v0 = gkey_vertex[TOL.geometric_key(v0_xyz)]
    #             v1 = gkey_vertex[TOL.geometric_key(v1_xyz)]
    #             names = ["is_quad", "u_edge", "v_edge", "brep_face"]
    #             values = [True, (u0, u1), (v0, v1), brep_face]
    #             subdmesh.face_attributes(face, names, values)

    #         else:
    #             for curve, edge in zip(segments, subdmesh.face_halfedges(face)):
    #                 sp = curve.PointAtStart
    #                 ep = curve.PointAtEnd
    #                 u = gkey_vertex[TOL.geometric_key(sp)]
    #                 v = gkey_vertex[TOL.geometric_key(ep)]
    #                 subdmesh.edge_attribute(edge, "brep_curve", curve)
    #                 subdmesh.edge_attribute(edge, "brep_curve_dir", (u, v))

    #     return subdmesh

    # ==========================================================================
    # topology
    # ==========================================================================

    def subd_edge_strip(self, edge):
        """
        Find the edge strip through quad and nonquad faces.

        Parameters
        ----------
        edge : tuple[int, int]
            The edge identifier.

        Returns
        -------
        list[tuple[int, int]]

        """

        def strip_end_faces(strip):
            # return nonquads at the end of edge strips
            faces1 = self.edge_faces(strip[0])
            faces2 = self.edge_faces(strip[-1])
            nonquads = []
            for face in faces1 + faces2:
                if face is not None and len(self.face_vertices(face)) != 4:
                    nonquads.append(face)
            return nonquads

        strip = self.edge_strip(edge)

        all_edges = list(strip)

        end_faces = set(strip_end_faces(strip))
        seen = set()

        while len(end_faces) > 0:
            face = end_faces.pop()
            if face not in seen:
                seen.add(face)
                for u, v in self.face_halfedges(face):
                    halfedge = (u, v)
                    if halfedge not in all_edges:
                        rev_hf_face = self.halfedge_face((v, u))
                        if rev_hf_face is not None:
                            if len(self.face_vertices(rev_hf_face)) != 4:
                                end_faces.add(self.halfedge_face((v, u)))
                                all_edges.append(halfedge)
                                continue
                        halfedge_strip = self.edge_strip(halfedge)
                        all_edges.extend(halfedge_strip)
                        end_faces.update(strip_end_faces(halfedge_strip))
        return all_edges

    def edge_strip_faces(self, edge_strip):
        """
        Identify all edge faces of the edge strip.

        Parameters
        ----------
        edge_strip : list[tuple[int, int]]
            A list of edges.

        Returns
        -------
        list[int]

        """
        edge_strip_faces = set()
        for edge in edge_strip:
            face1, face2 = self.edge_faces(edge)
            if face1 is not None:
                edge_strip_faces.add(face1)
            if face2 is not None:
                edge_strip_faces.add(face2)
        return list(edge_strip_faces)

    def split_boundary(self, subdmesh):
        # type: (SubdMesh) -> list[list[int]]
        """
        Split the ordered list of boundary vertices at the corners of the quadmesh.

        Parameters
        ----------
        subdmesh : :class:`SubdMesh`

        Returns
        -------
        list[list[int]]

        """
        boundaries = subdmesh.vertices_on_boundaries()
        exterior = boundaries[0]
        opening = []
        openings = [opening]
        for vertex in exterior:
            opening.append(vertex)
            if subdmesh.vertex_degree(vertex) == 2:
                opening = [vertex]
                openings.append(opening)
        openings[-1] += openings[0]
        del openings[0]
        openings[:] = [opening for opening in openings if len(opening) > 2]
        return openings

    # ==========================================================================
    # remapping
    # ==========================================================================

    # def divide_brep_curve(self, edge, n):
    #     """
    #     Divide the brep curve corresponding to an edge and store the resulting points in the edge attributes.

    #     Parameters
    #     ----------
    #     edge : tuple[int, int]
    #         The identifier of the edge.
    #     n : int
    #         The number of segments.

    #     Returns
    #     -------
    #     list[Rhino.Geometry.Point3d]

    #     """
    #     curve = self.edge_attribute(edge, "brep_curve")
    #     params = curve.DivideByCount(n, True)
    #     pts = []
    #     for param in params:
    #         pt = curve.PointAt(param)
    #         pts.append(pt)
    #     self.edge_attribute(edge, "brep_curve_pts", pts)
    #     return pts

    def remap_boundary_vertices(self, subdmesh):
        # type: (SubdMesh) -> SubdMesh
        """
        Remap the boundary vertices of a subdmesh.

        Parameters
        ----------
        subdmesh : :class:`SubdMesh`
            Another subd mesh.

        Returns
        -------
        :class:`SubdMesh`

        """
        vertex_loops = self.split_boundary(subdmesh)
        for vertex_loop in vertex_loops:
            vertex_loop = [key[0] for key in groupby(vertex_loop)]
            u = vertex_loop[0]
            v = vertex_loop[-1]
            edge = (u, v)
            pts = self.edge_attribute(edge, "brep_curve_pts")
            if pts:
                if u != self.edge_attribute(edge, "brep_curve_dir")[0]:
                    pts = pts[::-1]
                for pt, key in zip(pts[1:-1], vertex_loop[1:-1]):
                    subdmesh.vertex_attribute(key, "x", pt.X)
                    subdmesh.vertex_attribute(key, "y", pt.Y)
                    subdmesh.vertex_attribute(key, "z", pt.Z)
        return subdmesh

    # ==========================================================================
    # subdivision
    # ==========================================================================

    def subdivide_quad(self, face):
        """
        Subdivide a quad face using the underlying BRep face geometry.

        The number of subdivisions in the U and V directions
        depends on the parameters ``nu`` and ``nv`` stored in the face attributes.
        The geometry of the BRep face is also stored in the face attributes.

        Parameters
        ----------
        face : int
            The identifier of the face.

        Returns
        -------
        :class:`Mesh`

        """
        brep_face, nu, nv = self.face_attributes(face, names=["brep_face", "nu", "nv"])

        domain_u = brep_face.Domain(0)
        domain_v = brep_face.Domain(1)

        du = (domain_u[1] - domain_u[0]) / (nu)
        dv = (domain_v[1] - domain_v[0]) / (nv)

        gkey_vertex = {}
        subdmesh = Mesh()
        for vertex in self.face_vertices(face):
            x, y, z = self.vertex_coordinates(vertex)
            subdmesh.add_vertex(vertex, {"x": x, "y": y, "z": z})
            gkey_vertex[TOL.geometric_key((x, y, z))] = vertex

        point_at = brep_face.PointAt

        for i in range(nu):
            for j in range(nv):
                a = point_at(domain_u[0] + (i + 0) * du, domain_v[0] + (j + 0) * dv)
                b = point_at(domain_u[0] + (i + 1) * du, domain_v[0] + (j + 0) * dv)
                c = point_at(domain_u[0] + (i + 1) * du, domain_v[0] + (j + 1) * dv)
                d = point_at(domain_u[0] + (i + 0) * du, domain_v[0] + (j + 1) * dv)

                vertices = []
                for pt in [a, b, c, d]:
                    gkey = TOL.geometric_key(pt)
                    if gkey not in gkey_vertex:
                        gkey_vertex[gkey] = subdmesh.add_vertex(x=pt[0], y=pt[1], z=pt[2])
                    vertex = gkey_vertex[gkey]
                    vertices.append(vertex)
                subdmesh.add_face(vertices)

        return self.remap_boundary_vertices(subdmesh)

    def subdivide_nonquad(self, nonquad):
        """
        Apply recursive quad subdivision to a nonquad face using the underlying BRep face geometry.

        The number of subdivisions ``n`` is stored in the face attributes.
        The geometry of the BRep face is also stored in the face attributes.

        Parameters
        ----------
        nonquad : int
            The identifier of the face.

        Returns
        -------
        :class:`Mesh`

        """
        n = self.face_attribute(nonquad, "n")

        mesh = Mesh()
        vertices = self.face_vertices(nonquad)
        for vertex in vertices:
            x, y, z = self.vertex_coordinates(vertex)
            mesh.add_vertex(vertex, {"x": x, "y": y, "z": z})
        mesh.add_face(vertices)

        for _ in range(n):
            subd = mesh_fast_copy(mesh)  # type: Mesh

            edgepoints = []
            for u, v in mesh.edges():
                w = subd.split_edge((u, v), allow_boundary=True)
                edgepoints.append([w, True])

            face_xyz = {face: mesh.face_centroid(face) for face in mesh.faces()}

            for face in mesh.faces():
                vertex_descendant = {i: j for i, j in subd.face_halfedges(face)}
                vertex_ancestor = {j: i for i, j in subd.face_halfedges(face)}

                x, y, z = face_xyz[face]
                c = subd.add_vertex(x=x, y=y, z=z)

                for vertex in mesh.face_vertices(face):
                    a = vertex_ancestor[vertex]
                    d = vertex_descendant[vertex]
                    subd.add_face([a, vertex, d, c])

                # del subd.face[face]
                subd.delete_face(face)
            mesh = subd

        for edge in self.face_halfedges(nonquad):
            self.divide_brep_curve(edge, 2**n)

        subdmesh = Mesh.__from_data__(mesh.__data__)
        return self.remap_boundary_vertices(subdmesh)

    # def subdivide_all_faces(self):
    #     """
    #     Subdivide all the faces of the mesh, switching between a scheme based on the UV space of the underlying surface
    #     and a basic quad scheme, for quad and non-quad faces respectively.

    #     Returns
    #     -------
    #     :class:`Mesh`

    #     """
    #     subd_meshes = []
    #     quads = []
    #     non_quads = []

    #     boundary = set()

    #     for face in self.faces():
    #         if self.face_attribute(face, "is_quad"):
    #             quads.append(face)
    #         else:
    #             non_quads.append(face)

    #     for face in non_quads:
    #         subd_mesh = self.subdivide_nonquad(face)
    #         subd_meshes.append(subd_mesh)
    #         for vertex in subd_mesh.vertices_on_boundary():
    #             xyz = subd_mesh.vertex_coordinates(vertex)
    #             boundary.add(TOL.geometric_key(xyz))

    #     for face in quads:
    #         subd_mesh = self.subdivide_quad(face)
    #         subd_meshes.append(subd_mesh)
    #         for vertex in subd_mesh.vertices_on_boundary():
    #             xyz = subd_mesh.vertex_coordinates(vertex)
    #             boundary.add(TOL.geometric_key(xyz))

    #     mesh = meshes_join_and_weld(subd_meshes, precision="2f")

    #     fixed = []
    #     for vertex in mesh.vertices():
    #         xyz = mesh.vertex_coordinates(vertex)
    #         gkey = TOL.geometric_key(xyz)
    #         if gkey in boundary:
    #             fixed.append(vertex)

    #     mesh.smooth_area(fixed=fixed, kmax=100, damping=0.5)

    #     return mesh



from compas.plugins import plugin
from compas.scene.context import register

from compas_rv.datastructures import ForceDiagram
from compas_rv.datastructures import FormDiagram
from compas_rv.datastructures import Pattern
from compas_rv.datastructures import ThrustDiagram

from .forceobject import RhinoForceObject
from .formobject import RhinoFormObject
from .patternobject import RhinoPatternObject
from .thrustobject import RhinoThrustObject


@plugin(category="factories", pluggable_name="register_scene_objects", requires=["Rhino"])
def register_scene_objects_rhino():
    register(Pattern, RhinoPatternObject, context="Rhino")
    register(FormDiagram, RhinoFormObject, context="Rhino")
    register(ThrustDiagram, RhinoThrustObject, context="Rhino")
    register(ForceDiagram, RhinoForceObject, context="Rhino")



import Rhino  # type: ignore
import rhinoscriptsyntax as rs  # type: ignore
import scriptcontext as sc  # type: ignore

import compas_rhino.conversions
import compas_rhino.objects
from compas.colors import Color
from compas.scene.descriptors.color import ColorAttribute
from compas_rui.scene import RUIMeshObject
from compas_rv.datastructures import Diagram
from compas_rv.session import RVSession


class RhinoDiagramObject(RUIMeshObject):
    session = RVSession()

    freecolor = ColorAttribute(default=Color.white())
    anchorcolor = ColorAttribute(default=Color.red())
    fixedcolor = ColorAttribute(default=Color.cyan())

    def __init__(
        self,
        disjoint=True,
        show_supports=True,
        show_fixed=True,
        show_free=False,
        **kwargs,
    ):
        super().__init__(disjoint=disjoint, **kwargs)

        self.show_vertices = True
        self.show_supports = show_supports
        self.show_fixed = show_fixed
        self.show_free = show_free
        self.show_edges = True
        self.show_faces = False

        self._guids_angles = []

    # =============================================================================
    # Properties
    # =============================================================================

    @property
    def settings(self):
        settings = super().settings
        settings["show_supports"] = self.show_supports
        settings["show_fixed"] = self.show_fixed
        settings["show_free"] = self.show_free
        return settings

    @property
    def diagram(self) -> Diagram:
        return self.mesh

    @diagram.setter
    def diagram(self, diagram: Diagram) -> None:
        self.mesh = diagram

    def edges(self, **kwargs):
        return self.diagram.edges(**kwargs)

    def faces(self, **kwargs):
        return self.diagram.faces(**kwargs)

    def forces(self):
        return self.diagram.edges_attribute("_f", keys=self.edges())

    def compute_edge_colors(self, tol=1e-3) -> None:
        forces = self.forces()
        magnitudes = [abs(f) for f in forces]
        fmin = min(magnitudes)
        fmax = max(magnitudes)

        if fmax - fmin < tol:
            # size of the range of forces is already checked here
            # no need to check again in the loop
            return

        colors = []

        for force, magnitude in zip(forces, magnitudes):
            # this will need to be updated once we allow for tension forces
            # or we have to exclude tension forces from the calculation
            # and give tension edges their own style
            colors.append(Color.from_i((magnitude - fmin) / (fmax - fmin)))

        return colors

    # =============================================================================
    # Clear
    # =============================================================================

    def clear_angles(self):
        compas_rhino.objects.delete_objects(self._guids_angles, purge=True)

    def clear(self):
        super().clear()
        self.clear_angles()

    # =============================================================================
    # Draw
    # =============================================================================

    def draw(self):
        faces = []
        if self.show_faces:
            faces += list(self.diagram.faces_where(_is_loaded=True))
        if faces:
            self.show_faces = faces

        for vertex in self.diagram.vertices():
            if self.diagram.vertex_attribute(vertex, "is_support"):
                self.vertexcolor[vertex] = self.anchorcolor
            elif self.diagram.vertex_attribute(vertex, "is_fixed"):
                self.vertexcolor[vertex] = self.fixedcolor
            else:
                self.vertexcolor[vertex] = self.freecolor

        super().draw()

        if self.session.settings.drawing.show_angles:
            self.draw_angles()

        return self.guids

    def draw_vertices(self):
        if self.show_vertices is True:
            vertices = []
            if self.show_free:
                vertices += list(self.diagram.vertices_where(is_support=False, is_fixed=False))
            if self.show_fixed:
                vertices += list(self.diagram.vertices_where(is_fixed=True))
            if self.show_supports:
                vertices += list(self.diagram.vertices_where(is_support=True))
            self.show_vertices = vertices

        for vertex in self.diagram.vertices():
            if self.diagram.vertex_attribute(vertex, "is_support"):
                self.vertexcolor[vertex] = self.anchorcolor
            elif self.diagram.vertex_attribute(vertex, "is_fixed"):
                self.vertexcolor[vertex] = self.fixedcolor
            else:
                self.vertexcolor[vertex] = self.freecolor

        return super().draw_vertices()

    def draw_edges(self):
        if self.show_edges is True:
            edges = list(self.edges())
            if edges:
                self.show_edges = edges

        if self.session.settings.drawing.show_forces:
            edges = list(self.edges())
            colors = self.compute_edge_colors()
            if colors:
                edge_color = dict(zip(edges, colors))
            else:
                edge_color = dict()

        for edge in self.edges():
            if self.session.settings.drawing.show_forces:
                if edge in edge_color:
                    self.edgecolor[edge] = edge_color[edge]
            else:
                self.edgecolor.clear()  # not sure why this is not recognised

        return super().draw_edges()

    def draw_faces(self):
        faces = []
        if self.show_faces is True:
            faces += list(self.faces())
            if faces:
                self.show_faces = faces

        return super().draw_faces()

    def draw_angles(self):
        fontheight = 10
        fontface = "Arial Regular"
        group = None

        # perhaps there should be aseparate setting for this
        tol = self.session.settings.tna.horizontal_max_angle

        edges = list(self.edges())
        angles = self.diagram.edges_attribute(name="_a", keys=edges)
        amin = min(angles)
        amax = max(angles)
        aspan = amax - amin

        if aspan**2 < 0.001**2:
            return

        transformation = compas_rhino.conversions.transformation_to_rhino(self.worldtransformation)

        guids = []

        for edge, angle in zip(edges, angles):
            if angle < tol:
                continue

            name = "{}.angle.{}-{}".format(self.diagram.name, *edge)  # type: ignore
            color = Color.from_i((angle - amin) / aspan)
            attr = self.compile_attributes(name="{}.label".format(name), color=color)
            line = self.diagram.edge_line(edge)
            location = compas_rhino.conversions.point_to_rhino(line.midpoint)
            location.Transform(transformation)

            dot = Rhino.Geometry.TextDot(f"{angle:.1f}", location)  # type: ignore
            dot.FontHeight = fontheight
            dot.FontFace = fontface

            guid = sc.doc.Objects.AddTextDot(dot, attr)
            guids.append(guid)

        if group:
            self.add_to_group(group, guids)

        self._guids_angles = guids
        self._guids += guids
        return guids

    # =============================================================================
    # Redraw
    # =============================================================================

    def redraw(self):
        rs.EnableRedraw(False)
        self.clear()
        self.draw()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_vertices(self):
        rs.EnableRedraw(False)
        self.clear_vertices()
        self.draw_vertices()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_edges(self):
        rs.EnableRedraw(False)
        self.clear_edges()
        self.draw_edges()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_faces(self):
        rs.EnableRedraw(False)
        self.clear_faces()
        self.draw_faces()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_angles(self):
        rs.EnableRedraw(False)
        self.clear_angles()
        self.draw_angles()
        rs.EnableRedraw(True)
        rs.Redraw()



import rhinoscriptsyntax as rs  # type: ignore

from compas.colors import Color
from compas.scene.descriptors.color import ColorAttribute
from compas_rui.scene import RUIMeshObject
from compas_rv.datastructures import Pattern
from compas_rv.session import RVSession


class RhinoPatternObject(RUIMeshObject):
    session = RVSession()
    mesh: Pattern

    freecolor = ColorAttribute(default=Color.white())
    anchorcolor = ColorAttribute(default=Color.red())
    fixedcolor = ColorAttribute(default=Color.blue())

    def __init__(
        self,
        disjoint=True,
        vertexgroup="RV::FormDiagram::Vertices",
        edgegroup="RV::FormDiagram::Edges",
        facegroup="RV::FormDiagram::Faces",
        **kwargs,
    ):
        super().__init__(
            disjoint=disjoint,
            vertexgroup=vertexgroup,
            edgegroup=edgegroup,
            facegroup=facegroup,
            **kwargs,
        )

    # =============================================================================
    # Clear
    # =============================================================================

    # =============================================================================
    # Draw
    # =============================================================================

    def draw_vertices(self):
        for vertex in self.mesh.vertices():
            if self.mesh.vertex_attribute(vertex, "is_support"):
                self.vertexcolor[vertex] = self.anchorcolor
            elif self.mesh.vertex_attribute(vertex, "is_fixed"):
                self.vertexcolor[vertex] = self.fixedcolor
            else:
                self.vertexcolor[vertex] = self.freecolor

        return super().draw_vertices()

    # =============================================================================
    # Redraw
    # =============================================================================

    def redraw_vertices(self):
        rs.EnableRedraw(False)
        self.clear_vertices()
        self.draw_vertices()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_edges(self):
        rs.EnableRedraw(False)
        self.clear_edges()
        self.draw_edges()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw_faces(self):
        rs.EnableRedraw(False)
        self.clear_faces()
        self.draw_faces()
        rs.EnableRedraw(True)
        rs.Redraw()

    def redraw(self):
        rs.EnableRedraw(False)
        self.clear()
        self.draw()
        rs.EnableRedraw(True)
        rs.Redraw()



from compas.datastructures import Mesh
from compas.geometry import Line
from compas.geometry import angle_vectors
from compas.itertools import pairwise
from compas_fd.solvers import fd_numpy


class Pattern(Mesh):
    """
    Data structure for force layout patterns that are the basis for form diagrams.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.attributes.update({"openings": {}})
        self.default_vertex_attributes.update(
            {
                "x": 0.0,
                "y": 0.0,
                "z": 0.0,
                "is_fixed": False,
                "is_support": False,
            }
        )
        self.default_edge_attributes.update(
            {
                "q": 1.0,
                "lmin": 1e-6,
                "lmax": 1e6,
            }
        )

    def collapse_small_edges(self, tol=1e-2):
        """
        Collapse the edges that are shorter than a given length.

        Parameters
        ----------
        tol : float, optional
            The threshold length.
            Edges with a length shorter than this value
            will be collapsed.

        Returns
        -------
        None

        """
        for key in list(self.edges()):
            if self.has_edge(key):
                u, v = key
                if self.edge_length(u, v) < tol:
                    self.collapse_edge(u, v, t=0.5, allow_boundary=True)

    def smooth(self, fixed, kmax=10):
        """
        Apply area smoothing to the mesh geometry.

        Parameters
        ----------
        fixed : list[int]
            The vertices to keep fixed.
        kmax : int, optional
            The number of smoothing iterations.

        Returns
        -------
        None

        """
        self.smooth_area(self, fixed=fixed, kmax=kmax)

    def relax(self):
        """
        Relax the mesh using the force density method with the curent edge force densities.

        Returns
        -------
        None

        """
        vertex_index = self.vertex_index()
        xyz = self.vertices_attributes("xyz")
        loads = [[0.0, 0.0, 0.0] for _ in xyz]
        anchors = [vertex_index[key] for key in self.vertices_where(is_support=True)]
        fixed = [vertex_index[key] for key in self.vertices_where(is_fixed=True)]
        supports = list(set(anchors + fixed))
        edges = [(vertex_index[u], vertex_index[v]) for u, v in self.edges()]
        q = self.edges_attribute("q")
        result = fd_numpy(vertices=xyz, fixed=supports, edges=edges, forcedensities=q, loads=loads)
        for key in self.vertices():
            index = vertex_index[key]
            self.vertex_attributes(key, "xyz", result.vertices[index])

    def corner_vertices(self, tol=160):
        """
        Identify the corner vertices.

        Parameters
        ----------
        tol : float, optional
            The threshold value for the angle formed between two edges at a vertex
            for it to be considered a corner.
            Vertices with smaller angles are considered a corner.

        Returns
        -------
        list[int]

        """
        vkeys = []
        for key in self.vertices_on_boundary():
            if self.vertex_degree(key) == 2:
                vkeys.append(key)
            else:
                nbrs = []
                for nkey in self.vertex_neighbors(key):
                    if self.is_edge_on_boundary(key, nkey):
                        nbrs.append(nkey)
                u = self.edge_vector(key, nbrs[0])
                v = self.edge_vector(key, nbrs[1])
                if angle_vectors(u, v, deg=True) < tol:
                    vkeys.append(key)
        return vkeys

    def edge_loop_vertices(self, uv):
        """
        Identify all vertices on an edge loop.

        Parameters
        ----------
        uv : tuple[int, int]
            The identifier of the base edge of the loop.

        Returns
        -------
        list[int]

        """
        edges = self.edge_loop(uv)
        if len(edges) == 1:
            return edges[0]
        vertices = [edge[0] for edge in edges]
        if edges[-1][1] != edges[0][0]:
            vertices.append(edges[-1][1])
        return vertices

    def split_boundary(self):
        """Split the boundary into openings based on the location of the anchors.

        The algorithm assumes that the first boundary of the mesh is the exterior boundary.

        Returns
        -------
        list[list[int]]

        """
        boundaries = self.vertices_on_boundaries()
        exterior = boundaries[0]
        if exterior[-1] == exterior[0]:
            del exterior[-1]
        anchors = list(self.vertices_where(is_support=True))
        anchors = sorted(anchors, key=lambda a: exterior.index(a))
        index = exterior.index(anchors[0])
        exterior = exterior[index:] + exterior[:index]
        exterior.append(exterior[0])
        openings = []
        i = 0
        for anchor in anchors[1:]:
            j = exterior.index(anchor)
            openings.append(exterior[i : j + 1])
            i = j
        openings.append(exterior[j:])
        return [opening for opening in openings if len(opening) > 2]

    def compute_sag(self, opening):
        """Compute the sag of the opening as the ratio of the rise of the opening over its span.

        Parameters
        ----------
        opening : list[int]

        Returns
        -------
        float

        """
        start = self.vertex_point(opening[0])
        end = self.vertex_point(opening[-1])
        span = Line(start, end)
        distances = [self.vertex_point(vertex).distance_to_line(span) for vertex in opening[1:-1]]
        rise = max(distances)
        return rise / span.length

    def init_openings(self, minsag: float = 0.1) -> tuple[list[list[int]], list[float]]:
        """Initialise the boundary openings by imposing a minimal sag.

        Parameters
        ----------
        minsag : float, optional
            Minimal value for boundary sag.

        Returns
        -------
        tuple[list[list[int]], list[float]]

        """
        openings = self.split_boundary()

        targets = []
        for opening in openings:
            sag = self.compute_sag(opening)
            sag = max(sag, minsag)
            targets.append(sag)

        self.relax()
        self.match_opening_sag_targets(openings, targets)

        return openings, targets

    def match_opening_sag_targets(self, openings: list[list[int]], targets: list[float]) -> None:
        """Iteratively update the force densities in the opneing edges to match the target sag values.

        Parameters
        ----------
        openings : list[list[int]]
            The vertices of the openings.
        targets : list[float]
            The target sag values.

        """
        count = 0
        while count < 10:
            count += 1
            current = [self.compute_sag(opening) for opening in openings]

            if all(abs(sag - target) < 0.01 for sag, target in zip(current, targets)):
                break

            for sag, target, opening in zip(current, targets, openings):
                scale = sag / target

                for u, v in pairwise(opening):
                    q = self.edge_attribute((u, v), name="q")
                    self.edge_attribute((u, v), name="q", value=scale * q)

            self.relax()



from compas.geometry import scale_vector
from compas.geometry import sum_vectors
from compas_fd.solvers import fd_numpy
from compas_tna.diagrams import FormDiagram

from .diagram import Diagram
from .pattern import Pattern


class FormDiagram(Diagram, FormDiagram):
    """
    Data structure for form diagrams.
    """

    @classmethod
    def from_pattern(cls, pattern: Pattern) -> "FormDiagram":
        """Construct a form diagram from a pattern.

        Parameters
        ----------
        pattern : Pattern
            The pattern from which the diagram should be constructed.
        feet : {1, 2}, optional
            The number of horizontal force directions that should be added to the supports.

        Returns
        -------
        FormDiagram
            The form diagram.
        """
        form: FormDiagram = pattern.copy(cls=cls)
        form.update_boundaries()
        return form

    # not sure if this is a good idea
    # because it might clash with the parent function

    def edges_on_boundaries(self) -> list[list[tuple[int, int]]]:
        """Compute and return the edges on the perceived boundary of the diagram.

        Returns
        -------
        list[list[tuple[int, int]]]

        """
        boundaries = []
        for face in self.faces_where(_is_loaded=False):
            boundary = []
            for edge in self.face_halfedges(face):
                if self.edge_attribute(edge, name="_is_edge"):
                    boundary.append(edge)
            boundaries.append(boundary)
        return boundaries

    def is_vertex_internal(self, vertex: int) -> bool:
        """Indicate that a vertex is on perceived inside of the diagram.

        Parameters
        ----------
        vertex : int
            The identifier of the vertex.

        Returns
        -------
        bool

        """
        return not any(self.is_face_on_boundary(face) for face in self.vertex_faces(vertex))

    def solve_fd(self) -> None:
        """
        Relax the mesh using the force density method with the curent edge force densities.

        Returns
        -------
        None

        """
        vertex_index = self.vertex_index()
        xyz = self.vertices_attributes("xyz")
        loads = [[0.0, 0.0, 0.0] for _ in xyz]
        fixed = [vertex_index[key] for key in self.vertices_where(is_support=True)]
        fixed += [vertex_index[key] for key in self.vertices_where(is_fixed=True)]
        edges = list(self.edges_where(_is_edge=True))
        q = self.edges_attribute("q", keys=edges)
        edges = [(vertex_index[u], vertex_index[v]) for u, v in edges]
        result = fd_numpy(vertices=xyz, fixed=fixed, edges=edges, forcedensities=q, loads=loads)
        for key in self.vertices():
            index = vertex_index[key]
            self.vertex_attributes(key, "xyz", result.vertices[index])

    def flip_cycles_if_normal_down(self):
        """Flip the cycles of the diagram if the average normal points downward."""
        normals = [self.face_normal(face) for face in self.faces_where(_is_loaded=True)]
        scale = 1 / len(normals)
        normal = scale_vector(sum_vectors(normals), scale)
        if normal[2] < 0:
            self.flip_cycles()



import Rhino  # type: ignore
from numpy import array
from numpy import float64

import compas_rhino.conversions
from compas_fd.solvers.fd_numerical_data import FDNumericalData
from compas_rv.conduits import EdgesConduit
from compas_rv.datastructures import ThrustDiagram
from compas_tna.equilibrium.diagrams import update_z
from compas_tna.loads import LoadUpdater


class InteractiveScaleHorizontal:
    def __init__(
        self,
        thrust: ThrustDiagram,
    ):
        self.thrust = thrust
        self.scale = 1.0

        self._numdata = None
        self._loadupdater = None
        self._q0 = None
        self._conduit_edges = None

    def __call__(self):
        try:
            self.start()
            self.stop()
        except Exception as e:
            print(e)
            return False
        else:
            return True

    @property
    def conduit_edges(self) -> EdgesConduit:
        if not self._conduit_edges:
            self._conduit_edges = EdgesConduit(self.numdata.xyz, self.numdata.edges, thickness=1)
        return self._conduit_edges

    @property
    def numdata(self) -> FDNumericalData:
        if self._numdata is None:
            vertex_index = self.thrust.vertex_index()
            vertices = self.thrust.vertices_attributes("xyz")
            loads = [self.thrust.vertex_attributes(vertex, ["px", "py", "pz"]) or [0, 0, 0] for vertex in self.thrust.vertices()]
            fixed = [vertex_index[vertex] for vertex in self.thrust.vertices_where(is_support=True)]
            edges = list(self.thrust.edges_where(_is_edge=True))
            forcedensities = list(self.thrust.edges_attribute(name="q", keys=edges))
            edges = [(vertex_index[u], vertex_index[v]) for u, v in edges]
            self._numdata = FDNumericalData.from_params(vertices, fixed, edges, forcedensities, loads)
        return self._numdata

    # this can be done more elegantly
    @property
    def loadupdater(self) -> LoadUpdater:
        if self._loadupdater is None:
            self._loadupdater = LoadUpdater(
                self.thrust,
                array(self.numdata.p, copy=True),
                array(self.thrust.vertices_attribute("t"), dtype=float64).reshape((-1, 1)),
                1.0,
            )
        return self._loadupdater

    def _update(self) -> None:
        update_z(
            self.numdata.xyz,
            self.scale * self.numdata.Q,
            self.numdata.C,
            self.numdata.p,
            self.numdata.free,
            self.numdata.fixed,
            self.loadupdater,
            kmax=10,
        )

    def start(self):
        def on_dynamic_draw(sender, e):
            self.conduit_edges.disable()
            p2 = compas_rhino.conversions.point_to_compas(e.CurrentPoint)
            v02 = p2 - p0
            self.scale = v02.length / v01.length
            self._update()
            self.conduit_edges.enable()

        get_p0 = Rhino.Input.Custom.GetPoint()
        get_p0.SetCommandPrompt("Select base point")
        get_p0.Get()
        rhino_p0 = get_p0.Point()
        p0 = compas_rhino.conversions.point_to_compas(rhino_p0)

        get_p1 = Rhino.Input.Custom.GetPoint()
        get_p1.DrawLineFromPoint(rhino_p0, True)
        get_p1.SetCommandPrompt("Select reference 1")
        get_p1.Get()
        rhino_p1 = get_p1.Point()
        p1 = compas_rhino.conversions.point_to_compas(rhino_p1)

        v01 = p1 - p0

        get_p2 = Rhino.Input.Custom.GetPoint()
        get_p2.Constrain(rhino_p0, rhino_p1)
        get_p2.DrawLineFromPoint(rhino_p0, True)
        get_p2.SetCommandPrompt("Select reference 2")
        get_p2.DynamicDraw += on_dynamic_draw
        get_p2.Get()
        rhino_p2 = get_p2.Point()
        p2 = compas_rhino.conversions.point_to_compas(rhino_p2)

        v02 = p2 - p0

        self.scale = v02.length / v01.length  # scaling is always positive

    def stop(self):
        update_z(
            self.numdata.xyz,
            self.scale * self.numdata.Q,
            self.numdata.C,
            self.numdata.p,
            self.numdata.free,
            self.numdata.fixed,
            self.loadupdater,
            kmax=100,
        )

        self.conduit_edges.disable()
        del self._conduit_edges
        self._conduit_edges = None
